from typing import TypedDict, List, Literal, Optional, Union, Dict
from dataclasses import dataclass

from jorun.errors import TaskBuildException


class TaskOptions(TypedDict):
    pass


class ShellTask(TaskOptions):
    command: Union[str, List[str]]
    working_directory: Optional[str]
    environment: Optional[Dict[str, str]]


class DockerTask(TaskOptions):
    container_name: str
    image: str
    docker_arguments: Optional[List[str]]
    docker_command: Optional[List[str]]
    environment: Optional[Dict[str, str]]
    working_directory: Optional[str]
    stop_at_exit: bool


class Task(TypedDict):
    name: str
    type: Literal["shell", "docker", "group"]
    shell: Optional[ShellTask]
    docker: Optional[DockerTask]
    run_mode: Literal["wait_completion", "indefinite"]
    completion_pattern: Optional[str]
    pattern_in_stderr: Optional[bool]
    depends: Optional[List[str]]


class TasksConfiguration(TypedDict):
    tasks: Dict[str, Task]


@dataclass
class TaskNode:
    task: Task
    dependencies: List["TaskNode"]


def _setup_dependencies(parent_node: TaskNode, bucket: List[Task]):
    parent_node.dependencies = [
        TaskNode(task=t, dependencies=[]) for t in bucket if parent_node.task["name"] in (t.get("depends") or [])
    ]

    # remove dependencies from bucket
    for d in parent_node.dependencies:
        bucket.pop(bucket.index(d.task))

    # build dependencies for children
    for dep in parent_node.dependencies:
        _setup_dependencies(dep, bucket)


def build_task_tree(main_task_name: str, tasks: List[Task]) -> TaskNode:
    tasks_bucket = tasks.copy()

    main_task = next(t for t in tasks if t["name"] == main_task_name)

    if not main_task:
        raise TaskBuildException("No main task found")

    tasks_bucket.pop(tasks_bucket.index(main_task))

    # build deps tree
    tree = TaskNode(task=main_task, dependencies=[])
    _setup_dependencies(tree, tasks_bucket)

    return tree



import abc
from asyncio.subprocess import Process
from typing import Callable, Optional

from ..task import TaskOptions


class BaseTaskHandler(abc.ABC):
    @property
    @abc.abstractmethod
    def task_type(self) -> str:
        pass

    @abc.abstractmethod
    async def execute(self, options: Optional[TaskOptions], completion_callback: Callable, stderr_redirect: bool) \
            -> Optional[Process]:
        pass

    @abc.abstractmethod
    def on_exit(self, options: TaskOptions, process: Process):
        pass



import yaml
from .task import TasksConfiguration


def load_config(file_name: str) -> TasksConfiguration:
    with open(file_name, "r") as yamlf:
        config: TasksConfiguration = yaml.safe_load(yamlf)
        for t_name, t_task in config['tasks'].items():
            t_task['name'] = t_name

        return config



#!/usr/bin/env python
import argparse
import asyncio
import sys
from typing import List, Set, Dict

from .handler.group import GroupTaskHandler
from .handler.docker import DockerTaskHandler
from .handler.shell import ShellTaskHandler

from .configuration import load_config
from .runner import TaskRunner
from .task import Task
from .logger import logger

parser = argparse.ArgumentParser(prog="jorun", description="A smart task runner", add_help=True)

parser.add_argument("configuration_file", help="The yml configuration file to run")
parser.add_argument("--level", help="The log level (DEBUG, INFO, ...)", default="INFO", type=str)
parser.add_argument("--file-output", help="Log tasks output to files, one per task. "
                                          "This option lets you specify the directory of the log files", type=str)

running_tasks: List[TaskRunner] = []
async_tasks: Set[asyncio.Task] = set()

missing_tasks: Dict[str, Task] = {}
completed_tasks: Set[str] = set()

program_arguments: argparse.Namespace

task_handlers = [
    ShellTaskHandler(),
    DockerTaskHandler(),
    GroupTaskHandler()
]


def run_missing_tasks():
    tasks_to_run: List[Task] = [t for t in missing_tasks.values() if
                                (not t.get("depends") or len(t["depends"]) == 0) or set(t["depends"]).issubset(
                                    completed_tasks)]

    for task in tasks_to_run:
        missing_tasks.pop(task["name"])

    for task in tasks_to_run:
        run_task(task)


def run_task(task: Task):
    t = TaskRunner(task, task_handlers, program_arguments.file_output, program_arguments.level)
    running_tasks.append(t)

    def cb():
        logger.debug(f"Task {task['name']} completed")
        completed_tasks.add(task["name"])

        logger.debug(f"Launching task {task['name']} dependencies")
        run_missing_tasks()

    logger.debug(f"Running task {task['name']}")
    async_t = asyncio.get_event_loop().create_task(t.start(cb))
    async_tasks.add(async_t)
    async_t.add_done_callback(async_tasks.discard)


def main():
    global missing_tasks, program_arguments

    program_arguments = parser.parse_args()

    logger.setLevel(program_arguments.level)

    logger.debug("Loading configuration file")
    config = load_config(program_arguments.configuration_file)

    logger.debug("Building dependencies tree")
    # no dependency tree. we now support multiple dependencies
    # dep_tree = build_task_tree(config["main_task"], list(config["tasks"].values()))

    missing_tasks = config["tasks"].copy()

    loop = asyncio.get_event_loop()

    try:
        run_missing_tasks()
        loop.run_forever()
    except KeyboardInterrupt:
        for t in async_tasks:
            t.cancel()

        for i in reversed(range(len(running_tasks))):
            t = running_tasks[i]
            t.stop()
            running_tasks.pop(i)

        loop.stop()


if __name__ == "__main__":
    main()



import os
import subprocess
import asyncio
from asyncio.subprocess import Process
from typing import Callable, Optional

from ..task import TaskOptions, ShellTask
from .base import BaseTaskHandler
from ..logger import logger


class ShellTaskHandler(BaseTaskHandler):
    @property
    def task_type(self) -> str:
        return "shell"

    async def execute(self, options: Optional[ShellTask], completion_callback: Callable, stderr_redirect: bool) \
            -> Optional[Process]:
        stderr_file = subprocess.STDOUT if stderr_redirect else subprocess.PIPE

        out_cmd = options['command']
        if isinstance(out_cmd, list):
            out_cmd = " ".join(out_cmd)

        logger.debug(f"Running command: {out_cmd}")
        envs = options.get("environment")
        env_vars = None

        if envs:
            env_vars = dict(os.environ)
            env_vars.update({k: str(v) for k, v in envs.items()})

        if not isinstance(options["command"], list):
            process = await asyncio.create_subprocess_shell(
                options["command"],
                cwd=options.get("working_directory"),
                env=env_vars,
                stdout=subprocess.PIPE,
                stderr=stderr_file,
                stdin=subprocess.DEVNULL)
        else:
            process = await asyncio.create_subprocess_exec(
                options["command"][0],
                *options["command"][1:],
                cwd=options.get("working_directory"),
                env=env_vars,
                stdout=subprocess.PIPE,
                stderr=stderr_file,
                stdin=subprocess.DEVNULL)

        return process

    def on_exit(self, options: TaskOptions, process: Process):
        pass



import asyncio
import logging
import re
from asyncio.subprocess import Process
from typing import Callable, Optional

from .errors import TaskRunException


class AsyncScanner:
    _read_timeout = 1
    _process: Process
    _completion_callback: Optional[Callable]
    _stderr_print: bool
    _logger: logging.Logger
    _err_logger: logging.Logger

    def __init__(self, logger: logging.Logger, err_logger: logging.Logger, process: Process,
                 completion_callback: Callable, print_stderr: bool = False):
        self._process = process
        self._completion_callback = completion_callback
        self._stderr_print = print_stderr
        self._logger = logger
        self._err_logger = err_logger

    @staticmethod
    async def _readline(stream: asyncio.StreamReader, timeout: float):
        try:
            return await asyncio.wait_for(stream.readline(), timeout=timeout)
        except asyncio.TimeoutError:
            return None

    async def print_and_scan(self, pattern: str, task_name: str = "unknown"):
        if self._stderr_print:
            return await asyncio.gather(self._print_and_scan_stdout(pattern, task_name), self._print_stderr(task_name))
        else:
            return await self._print_and_scan_stdout(pattern, task_name)

    async def print(self, task_name: str = "unknown"):
        if self._stderr_print:
            return await asyncio.gather(self._print_stdout(task_name), self._print_stderr(task_name))
        else:
            return await self._print_stdout(task_name)

    async def _print_and_scan_stdout(self, pattern: str, task_name: str):
        reg = re.compile(pattern)
        pattern_matched = False

        try:
            while self._process.returncode is None:
                line_b = await self._readline(self._process.stdout, self._read_timeout)

                if line_b is not None:
                    line = line_b.decode('utf-8', errors='ignore')
                    self._logger.info(line, extra={'subprocess': task_name})
                    if reg.match(line) and self._completion_callback:
                        pattern_matched = True

                        self._completion_callback()
                        self._completion_callback = None
                        return await self._print_stdout(task_name)

            if not pattern_matched:
                while line_b := await self._process.stdout.readline():
                    line = line_b.decode('utf-8', errors='ignore')
                    self._logger.info(line, extra={'subprocess': task_name})
                    if reg.match(line) and self._completion_callback:
                        pattern_matched = True

                        self._completion_callback()
                        self._completion_callback = None
                        return await self._print_stdout(task_name)
        except Exception:
            pass

        if not pattern_matched:
            raise TaskRunException(f"Could not match given pattern on '{task_name}' before process exit")

    async def _print_stdout(self, task_name: str):
        try:
            while self._process.returncode is None:
                line_b = await self._readline(self._process.stdout, self._read_timeout)

                if line_b is not None:
                    self._logger.info(line_b.decode('utf-8', errors='ignore'), extra={'subprocess': task_name})

            while line_b := await self._process.stdout.readline():
                self._logger.info(line_b.decode('utf-8', errors='ignore'), extra={'subprocess': task_name})

            if self._completion_callback:
                self._completion_callback()
                self._completion_callback = None
        except:
            pass

    async def _print_stderr(self, task_name: str):
        try:
            while self._process.returncode is None:
                line_b = await self._readline(self._process.stderr, self._read_timeout)

                if line_b is not None:
                    self._err_logger.info(line_b.decode('utf-8', errors='ignore'), extra={'subprocess': task_name})

            while line_b := await self._process.stderr.readline():
                self._err_logger.info(line_b.decode('utf-8', errors='ignore'), extra={'subprocess': task_name})
        except:
            pass



import subprocess
from asyncio.subprocess import Process
from typing import Callable, Optional

from ..handler.base import BaseTaskHandler
from ..task import TaskOptions


class GroupTaskHandler(BaseTaskHandler):
    @property
    def task_type(self) -> str:
        return "group"

    async def execute(self, options: Optional[TaskOptions], completion_callback: Callable, stderr_redirect: bool) \
            -> Optional[Process]:
        completion_callback()
        return None

    def on_exit(self, options: TaskOptions, process: Process):
        pass



import asyncio
import subprocess
from asyncio.subprocess import Process
from typing import Callable, Optional

from ..handler.base import BaseTaskHandler
from ..logger import logger
from ..task import DockerTask


class DockerTaskHandler(BaseTaskHandler):
    _stop_on_exit: bool

    def __init__(self):
        self._stop_on_exit = False

    @property
    def task_type(self) -> str:
        return "docker"

    async def execute(self, options: Optional[DockerTask], completion_callback: Callable, stderr_redirect: bool) \
            -> Optional[Process]:
        command = ["docker", "run", "--name", options["container_name"],
                   *(options.get("docker_arguments") or [])]

        for env_key, env_value in (options.get("environment") or {}).items():
            env_value_s = str(env_value).replace('"', '\\"')
            command.append("-e")
            command.append(f'{env_key}={env_value_s}')

        command.append(options["image"])
        command.extend(options.get("docker_command") or [])

        stderr_file = subprocess.STDOUT if stderr_redirect else subprocess.PIPE

        logger.debug(f"Running command: {' '.join(command)}")

        process = await asyncio.create_subprocess_exec(
            command[0],
            *command[1:],
            cwd=options.get("working_directory"),
            stdout=subprocess.PIPE,
            stderr=stderr_file,
            stdin=subprocess.DEVNULL)

        self._stop_on_exit = options.get("stop_at_exit", False)
        return process

    def on_exit(self, options: DockerTask, process: Process):
        if self._stop_on_exit:
            subprocess.run([
                "docker",
                "stop",
                options['container_name']
            ])



import asyncio
import logging
import os.path
import sys
from asyncio.subprocess import Process
from datetime import datetime
from logging import Logger
from typing import Optional, Callable, List, Union

from .handler.base import BaseTaskHandler
from .logger import NewlineStreamHandler
from .scanner import AsyncScanner
from .task import TaskOptions, Task

OUTPUT_READ_INTERVAL = 0.015


class TaskRunner:
    _handlers: List[BaseTaskHandler]
    _handler: BaseTaskHandler

    _task: Task
    _process: Optional[Process]
    _completion_callback: Optional[Callable]
    _scanner: AsyncScanner

    _logger: Logger
    _err_logger: Logger

    def __init__(self, task: Task, handlers: List[BaseTaskHandler], file_output_dir: Optional[str],
                 log_level: Union[int, str]):
        self._task = task
        self._handlers = handlers
        self._process = None
        self._running = True
        self._completion_callback = None

        self._handler = next(h for h in self._handlers if h.task_type == task['type'])

        if not self._handler:
            raise RuntimeError(f"Task type '{task['type']}' unrecognized")

        self._logger = logging.Logger(task["name"])
        self._logger.setLevel(log_level)
        self._err_logger = logging.Logger(task["name"])
        self._err_logger.setLevel(log_level)

        console_handler = NewlineStreamHandler(sys.stdout)
        console_handler.setFormatter(logging.Formatter("[%(subprocess)s]: %(message)s"))

        self._logger.addHandler(console_handler)

        console_handler_err = NewlineStreamHandler(sys.stderr)
        console_handler.setFormatter(logging.Formatter("[%(subprocess)s]: %(message)s"))

        self._err_logger.addHandler(console_handler_err)

        if file_output_dir and os.path.isdir(file_output_dir):
            now_time = datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
            output_file = os.path.join(file_output_dir, f"{task['name']}_{now_time}.log")
            file_handler = logging.FileHandler(output_file)
            file_handler.terminator = ''

            self._logger.addHandler(file_handler)
            self._err_logger.addHandler(file_handler)

    def _on_stop(self):
        self._handler.on_exit(self._task[self._handler.task_type], self._process)

    def stop(self):
        if self._process and self._process.returncode is None:
            self._on_stop()
            self._process.terminate()

    async def start(self, completion_callback: Callable):
        try:
            self._completion_callback = completion_callback
            t = self._task

            stderr_redirect = t.get('pattern_in_stderr', False)

            task_options: Optional[TaskOptions] = t.get(self._handler.task_type)

            self._process = await self._handler.execute(task_options, self._completion_callback,
                                                        stderr_redirect)
            if not self._process:
                self._running = False
                return

            run_mode = t.get("run_mode") or "await_completion"
            completion_pattern = t.get("completion_pattern")
            self._scanner = AsyncScanner(self._logger, self._err_logger, self._process, self._completion_callback,
                                         not stderr_redirect)

            if run_mode == "await_completion" and completion_pattern:
                await self._scanner.print_and_scan(completion_pattern, self._task['name'])
            else:
                await self._scanner.print(self._task['name'])
        except asyncio.CancelledError:
            pass



import logging
import sys


class NewlineStreamHandler(logging.StreamHandler):
    def emit(self, record):
        try:
            msg = self.format(record)
            stream = self.stream
            # issue 35046: merged two stream.writes into one.
            final_message = msg if msg.endswith(self.terminator) else msg + self.terminator
            stream.write(final_message)
            self.flush()
        except RecursionError:  # See issue 36272
            raise
        except Exception:
            self.handleError(record)


logger = logging.Logger("runsk")
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logging.Formatter("::[%(levelname)s]: %(message)s"))
logger.addHandler(handler)
